using System;
using System.IO;
using System.Drawing;
using System.Threading;
using System.Collections.Generic;
using SdlDotNet;
using System.Runtime.InteropServices;

using Physics.States;
using Physics.Entities;
using Physics.Config.Files;

namespace Physics
{
    /* the Main class of our Physics system
     * creates everything and runs the main loop
     */
    static class Game
    {
        // the entrance to our program
        // [STAThread] // needed for the map open/save dialogs ...
        public static void Main()
        {
            Initalize();
            Run();
        }

        // holds the Surface that will be updated to the Computer Screen
        private static Surface screen;

        private static bool shutingDown;

        // a list of the game states
        static List<GameState> States;

        // holds the size of the video output 'port'
        public static Point videoSize = new Point(800, 600);

        /* Values for the Main Loop
         * TARGET_UPS := the amount of updates that is wanted in one second
         * ups := updates per second for the last second
         * fps := frames per second for the last second
         * updateCount := counts this seconds updates
         * frameCount  := counts this seconds draws
         */
        static int target_ups = 150; // a moving target that tries to maintain the fps
        const int MIN_UPS = 60;
        const int MAX_UPS = 200;
        const int TARGET_FPS = 100;
        static float updateSum = 0;

        static long lastUpdate;
        static long lastSecond;

        static int ups, fps;
        static int updateCount, drawCount;

        static int framesToAvg = 10;
        static long runStart;
        static long lastUpdateTook;
        static long lastDrawTook;
        static long lastRunTook;
        static float rollingUpdateTime = 1000;
        static float rollingDrawTime = 1000;
        static float rollingRunTime = 1000;
        static float rollingOtherTime = 1000;

        private static Global_cfg cfg;

        public static int getFPS
        {
            get { return fps; }
        } // used to get info to the debuger
        public static int getUPS
        {
            get { return ups; }
        }

        // sets up the basic game world for play
        static void Initalize()
        {
            // set screen to the Video Surface that will be displayed
            screen = Video.SetVideoModeWindow(videoSize.X, videoSize.Y, 24);
            Video.WindowCaption = "Basic Collisions : V1";

            States = new List<GameState>();
            AddGameState(new Running());

            Events.Quit += new QuitEventHandler(Events_Quit);

            // Adds any Entities to the world that exist at the start
            EntityCreator.StartEntities();

            // Adds Effects that exist at the start
            EffectsManager.Initalize();

            Debugger.Initalize();

            cfg = new Global_cfg();
        }

        static void Events_Quit(object sender, QuitEventArgs e)
        {
            shutingDown = true;
        }

        // the main game loop
        static void Run()
        {
            shutingDown = false;

            lastUpdate = TickCount;
            lastSecond = TickCount;

            while (!shutingDown)
            {
                DrawWorld();

                // Update game only when needed
                updateSum += (TickCount - lastUpdate) / (1000 / (float)target_ups);
                for (int i = 0; i < (int)updateSum; i++)
                    UpdateWorld();

                updateSum -= (int)updateSum;

                // Check if a second has passed to recalculate UPS and FPS
                if (TickCount - lastSecond >= 1000)
                {
                    lastSecond = TickCount;
                    ups = updateCount;
                    fps = drawCount;
                    updateCount = 0;
                    drawCount = 0;

                    // setUPStarget();
                }

                lastRunTook = TickCountMS - runStart;
                runStart = TickCountMS;
                rollingRunTime = RollingAverage(rollingRunTime, lastRunTook);

                rollingOtherTime = (rollingRunTime - rollingUpdateTime * updateSum - rollingDrawTime);
            }
        }

        // needs a bit of work
        private static void setUPStarget()
        {
            if (ups > 0)
                target_ups += (int)((fps - TARGET_FPS) * fps / ups);

            if (target_ups > MAX_UPS)
                target_ups = MAX_UPS;
            if (target_ups < MIN_UPS)
                target_ups = MIN_UPS;
        }

        private static float RollingAverage(float oldAverage, long lastTime)
        {
            return (oldAverage * (framesToAvg - 1) + lastTime) / framesToAvg;
        }

        // Handles Input
        static void Input()
        {
            InputMonitor.UpdateStates();

            if (cfg.exit.wasPressed())
                shutingDown = true;

            for (int i = 0; i < States.Count - 1; i++)
                States[i].Input(false);
            States[States.Count - 1].Input(true);
        }

        static void UpdateWorld()
        {
            lastUpdateTook = TickCountMS;

            Physics.Entities.Physics.CollisionHandler.CollisionChecks = 0;
            Physics.Entities.Physics.CollisionHandler.ValidCollisions = 0;

            Input();
            Update();

            updateCount += 1;

            lastUpdateTook = TickCountMS - lastUpdateTook;
            rollingUpdateTime = RollingAverage(rollingUpdateTime, lastUpdateTook);
        }

        // Update all Objects in the game world
        static void Update()
        {
            float timeStep = (TickCount - lastUpdate) / 8F;

            Events.Poll();  // polls through any events that have occured to the window
                            // includes things such as keyboard/mouse input and window movement.

            for (int i = 0; i < States.Count - 1; i++)
                States[i].Update(timeStep);
            States[States.Count - 1].Update(timeStep, true);

            lastUpdate = TickCount;
        }

        static void DrawWorld()
        {
            lastDrawTook = TickCountMS;
            Draw();
            drawCount += 1;
            lastDrawTook = TickCountMS - lastDrawTook;
            rollingDrawTime = RollingAverage(rollingDrawTime, lastDrawTook);
        }

        // Draw all objects to screen, also delay to reduce framerate
        static void Draw()
        {
            // clears the screen (background draw)
            screen.Fill(Color.FloralWhite);

            for(int i = 0; i < States.Count - 1; i++)
                States[i].Draw(screen, false);
            States[States.Count - 1].Draw(screen, true);

            // draws any Debug information to the screen
            Debugger.Draw(screen);

            // shows the screen that is in memmory to the visible screen
            screen.Flip();
        }

        public static bool RemoveGameState(GameState gState)
        {
            if (States[States.Count - 1] == gState)
            {
                States.RemoveAt(States.Count - 1);
                InputMonitor.Pop();
                return true;
            }
            else
            {
                Console.WriteLine("WARNING: a state wanted te be removed, but it was no onTop");
                return false;
            }
        }
        public static void AddGameState(GameState gState)
		{
            if (States.Contains(gState))
                Console.WriteLine("WARNING: a state was added that was already present!");

            InputMonitor.Push();
			States.Add(gState);
        }

		
        #region High Accuracy Timers
		
        // accurate to millisecond
        public static long TickCount
        {
            get
            {
                return DateTime.Now.Ticks / 10000;
            }
        }

        // accurate to microseconds
        public static long TickCountMS
        {
            get
            {
			    return DateTime.Now.Ticks / 10;
            }
        }

        /*
        [DllImport("kernel32.dll")]
        public static extern int QueryPerformanceFrequency (ref Int64 lpFrequency);
        [DllImport("kernel32.dll")]
        public static extern int QueryPerformanceCounter (ref Int64 lpPerformanceCount);
        
        System.Int64 freq = 1;
        System.Int64 tickcount = 0;

        QueryPerformanceFrequency(ref freq);
        QueryPerformanceCounter(ref tickcount);

        return (int)(tickcount * 1000000 / freq);
         */

        #endregion

        private static void testing()
        {
            float num = (1000000 - TARGET_FPS * (rollingDrawTime + rollingOtherTime)) / rollingUpdateTime / TARGET_FPS;
            int fps2 = (int)(1000000 / (rollingDrawTime + rollingOtherTime));
            int ups2 = (int)(1000000 / (rollingUpdateTime + rollingOtherTime));
            Console.WriteLine("FPS: " + fps2 + "  UPS: " + ups2);
            Console.WriteLine("rollingOtherTime: " + rollingOtherTime);
            Console.WriteLine("num: " + num);
            Console.WriteLine();
        }
    }
}
